from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.forms import ModelForm, fields
from .models import User


class UserForm(ModelForm):

    language_code = fields.ChoiceField(
        label=_('Language'),
        choices=settings.LANGUAGES,
        required=True
    )

    class Meta:
        model = User
        fields = [
            'username',
            'first_name',
            'last_name',
        ]

    def save(self, commit=True):
        super().save(commit=False)
        self.instance.language_code = self.cleaned_data['language_code']
        if commit:
            self.instance.save()
            self.save_m2m()



import logging
from celery import shared_task

from django.core import mail
from django.db import transaction
from django.utils.html import strip_tags

from .models import SystemMail

_logger = logging.getLogger(__name__)


@shared_task()
def run_mail_queue():
    with transaction.atomic():
        _logger.info('Starting System Mail Queue')
        mails_to_send = SystemMail.objects.filter(
            sent=False).select_for_update()
        for email in mails_to_send:
            _logger.info(f'Sending mail: {email}')
            try:
                msg = strip_tags(email.body)
                mail.send_mail(
                    email.subject, msg, email.from_name,
                    email.to_addr.split(','), html_message=email.body
                )
            except Exception as e:
                error_str = str(e)
                _logger.error(f'Failed to send system mail\n{error_str}')
                email.error = error_str
                email.save()
                continue

            email.sent = True
            email.error = None
            email.save()



from dataclasses import dataclass, field
from enum import Enum


class TableFieldAlignment(Enum):

    LEFT = 'left'
    CENTER = 'center'
    RIGHT = 'right'


class TableFieldType(Enum):

    NONE = ''
    STRING = '_string'
    MONETARY = '_monetary'
    FLOAT = '_float'


@dataclass
class TableField:

    label: str
    name: str
    alignment: TableFieldAlignment | Enum = TableFieldAlignment.LEFT
    header_alignment: TableFieldAlignment | Enum = None
    header_info: str = None
    field_type: TableFieldType | Enum = TableFieldType.NONE
    prefix: str = ''
    suffix: str = ''
    truncate_after: int = 0
    template: str = None


@dataclass
class BreadCrumb:

    name: str
    url: str


@dataclass
class ClientAction:

    name: str
    url: str = ''
    query_params: str = ''
    attrs: list[tuple[str, str]] = field(default_factory=list)
    submit: bool = False
    form_id: str = 'form'
    class_list: list[str] = field(default_factory=list)



import calendar
from datetime import datetime, date, timedelta
from dateutil.relativedelta import relativedelta
from django.utils import timezone


def current_year_start() -> datetime:
    now = timezone.now()
    first_day = now.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )
    # calendar.monthrange(date.year, date.month)[1]
    return first_day


def current_year_end() -> datetime:
    now = timezone.now()
    last_day = now.replace(
        month=12, day=31, hour=23, minute=59, second=59, microsecond=999999
    )
    return last_day


def current_month_start() -> datetime:
    now = timezone.now()
    monthrange = calendar.monthrange(now.year, now.month)
    first_day = now.replace(
        day=1, hour=0, minute=0, second=0, microsecond=0
    )
    return first_day


def current_month_end() -> datetime:
    now = timezone.now()
    monthrange = calendar.monthrange(now.year, now.month)
    last_day = now.replace(
        day=monthrange[1], hour=23, minute=59, second=59, microsecond=999999
    )
    return last_day



# Generated by Django 4.2.6 on 2023-10-10 20:06

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('accrete', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='member',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='memberships', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='accessgroupmember',
            name='access_group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accrete.accessgroup'),
        ),
        migrations.AddField(
            model_name='accessgroupmember',
            name='member',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accrete.member'),
        ),
        migrations.AddConstraint(
            model_name='accessgroup',
            constraint=models.UniqueConstraint(fields=('code',), name='unique_code'),
        ),
        migrations.AddConstraint(
            model_name='tenantaccessgroup',
            constraint=models.UniqueConstraint(fields=('tenant', 'access_group'), name='unique_tenant_per_group'),
        ),
        migrations.AddConstraint(
            model_name='accessgroupmember',
            constraint=models.UniqueConstraint(fields=('member', 'access_group'), name='unique_member_per_group'),
        ),
    ]



from django.contrib import admin
from . import models


class MemberInLine(admin.TabularInline):

    model = models.Member


class TenantAccessGroupInLine(admin.TabularInline):

    model = models.TenantAccessGroup


class AccessGroupMemberInLine(admin.TabularInline):

    model = models.AccessGroupMember


class TenantAdmin(admin.ModelAdmin):

    model = models.Tenant
    list_display = ('name', 'is_active', 'pk')
    search_fields = ('pk', 'name')
    list_filter = ['is_active']
    inlines = [MemberInLine, TenantAccessGroupInLine]


class MemberAdmin(admin.ModelAdmin):

    model = models.Member
    list_display = ('user', 'tenant', 'is_active')
    search_fields = ('user__email', 'tenant__name')
    list_filter = ['is_active']
    inlines = [AccessGroupMemberInLine]


class AccessGroupAdmin(admin.ModelAdmin):

    model = models.AccessGroup
    list_display = ('name', 'code')
    search_fields = ('name', 'code')
    inlines = [AccessGroupMemberInLine]


admin.site.register(models.Tenant, TenantAdmin)
admin.site.register(models.Member, MemberAdmin)
admin.site.register(models.AccessGroup, AccessGroupAdmin)
admin.site.register(models.AccessGroupMember)



from .filter import Filter
from .context import (
    ListContext,
    DetailContext,
    FormContext,
)
from .components import (
    ClientAction,
    BreadCrumb,
    TableField,
    TableFieldAlignment,
    TableFieldType
)
from .querystring import load_querystring, build_url_params



from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.validators import UnicodeUsernameValidator
from accrete.tenant import get_tenant


class TenantManager(models.Manager):

    def get_queryset(self):
        queryset = super().get_queryset()
        tenant = get_tenant()
        if tenant:
            queryset = queryset.filter(tenant=tenant)
        return queryset

    def bulk_create(
            self,
            objs,
            batch_size=None,
            ignore_conflicts=False,
            update_conflicts=False,
            update_fields=None,
            unique_fields=None,
    ):
        tenant = get_tenant()
        if tenant is None and not all(obj.tenant_id for obj in objs):
            raise ValueError('Tenant must be set for all objects when calling bulk_create')
        else:
            for obj in objs:
                obj.tenant_id = tenant.pk
        return super().bulk_create(
            objs, batch_size=batch_size, ignore_conflicts=ignore_conflicts,
            update_conflicts=update_conflicts, update_fields=update_fields,
            unique_fields=unique_fields
        )


class TenantModel(models.Model):

    class Meta:
        abstract = True

    tenant = models.ForeignKey(
        to='accrete.Tenant',
        on_delete=models.CASCADE
    )

    objects = TenantManager()

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):

        tenant = get_tenant()
        if tenant is not None and self.tenant != tenant:
            raise ValueError('Current tenant differs from tenant of the record!')
        self.tenant = self.tenant or tenant
        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields
        )


class Tenant(models.Model):

    class Meta:
        verbose_name = _('Tenant')
        verbose_name_plural = _('Tenants')
        ordering = ['-is_active', 'name']
        db_table = 'accrete_tenant'

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=255
    )

    is_active = models.BooleanField(
        verbose_name=_('Active'),
        default=True
    )

    access_groups = models.ManyToManyField(
        to='accrete.AccessGroup',
        through='accrete.TenantAccessGroup',
        through_fields=('tenant', 'access_group')
    )

    def __str__(self):
        return self.name


class Member(models.Model):

    class Meta:
        verbose_name = _('Member')
        verbose_name_plural = _('Members')
        ordering = ['tenant', 'user']
        db_table = 'accrete_member'

    username_validator = UnicodeUsernameValidator()

    user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        related_name='memberships',
        on_delete=models.PROTECT
    )

    tenant = models.ForeignKey(
        to='accrete.Tenant',
        related_name='members',
        on_delete=models.CASCADE
    )

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=150,
        help_text=_(
            '150 characters or fewer.'
            'Letters, digits and @/./+/-/_ only.'
        ),
        blank=True,
        null=True,
        validators=[username_validator],
    )

    is_active = models.BooleanField(
        verbose_name=_('Active'),
        default=True
    )

    access_groups = models.ManyToManyField(
        to='accrete.AccessGroup',
        through='accrete.AccessGroupMember',
        through_fields=('member', 'access_group')
    )

    objects = TenantManager()

    def __str__(self):
        return f'{self.name or self.user}'


class AccessGroup(models.Model):

    class Meta:
        verbose_name = _('Access Group')
        verbose_name_plural = _('Access Groups')
        ordering = ['name']
        db_table = 'accrete_access_group'
        constraints = [
            models.UniqueConstraint(
                name='unique_code',
                fields=['code']
            )
        ]

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=255
    )

    code = models.CharField(
        verbose_name=_('Code'),
        max_length=50
    )

    is_public = models.BooleanField(
        verbose_name=_('Is Public'),
        default=True,
        help_text=_(
            'If set, members from all tenants can be assigned to this group.'
        )
    )

    def __str__(self):
        return self.name


class AccessGroupMember(models.Model):

    class Meta:
        verbose_name = _('Access Group Member')
        verbose_name_plural = _('Access Group Members')
        ordering = ['member']
        db_table = 'accrete_access_group_member_rel'
        constraints = [
            models.UniqueConstraint(
                name='unique_member_per_group',
                fields=['member', 'access_group']
            )
        ]

    member = models.ForeignKey(
        to='accrete.Member',
        on_delete=models.CASCADE
    )

    access_group = models.ForeignKey(
        to='accrete.AccessGroup',
        on_delete=models.CASCADE
    )

    def __str__(self):
        return f'{self.member} - {self.access_group}'


class TenantAccessGroup(models.Model):

    class Meta:
        verbose_name = _('Tenant Access Group')
        verbose_name_plural = _('Tenant Access Groups')
        ordering = ['tenant']
        db_table = 'accrete_tenant_access_group_rel'
        constraints = [
            models.UniqueConstraint(
                name='unique_tenant_per_group',
                fields=['tenant', 'access_group']
            )
        ]

    tenant = models.ForeignKey(
        to='accrete.Tenant',
        on_delete=models.CASCADE
    )

    access_group = models.ForeignKey(
        to='accrete.AccessGroup',
        on_delete=models.CASCADE
    )

    def __str__(self):
        return f'{self.tenant} - {self.access_group}'



from django.views.generic import View
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth import logout, views
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect, render, reverse
from django.utils.translation import gettext_lazy as _

from accrete.forms import save_form
from accrete.contrib.ui import FormContext, DetailContext, ClientAction
from .forms import UserForm


class LoginView(views.LoginView):
    form_class = AuthenticationForm
    template_name = 'user/login.html'
    redirect_authenticated_user = True

    def form_invalid(self, form):
        user = form.get_user()
        if user is not None and not user.is_active:
            ctx = {'to_confirm': True}
            if self.extra_context:
                self.extra_context.update(ctx)
            else:
                self.extra_context = ctx
        return super().form_invalid(form)


class LogoutView(View):
    @staticmethod
    def get(request, *args, **kwargs):
        logout(request)
        return redirect('user:login')


@login_required()
def user_detail(request):
    ctx = DetailContext(
        request.user, request.GET.dict(), paginate_by=0, breadcrumbs=[],
        actions=[
            ClientAction(_('Edit'), url=reverse('user:edit')),
            ClientAction(_('Change E-Mail'), url=''),
            ClientAction(_('Change Password'), url='')
        ]
    )
    return render(request, 'user/user_detail.html', ctx.dict())


@login_required()
def user_edit(request):
    form = UserForm(
        initial={'language_code': request.user.language_code},
        instance=request.user
    )
    if request.method == 'POST':
        form = save_form(UserForm(request.POST, instance=request.user))
        if form.is_saved:
            return redirect('user:detail')
    ctx = FormContext(request.user, request.GET.dict(), context={'form': form}).dict()
    return render(request, 'user/user_form.html', ctx)



# Generated by Django 4.2.6 on 2023-10-10 20:06

import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AccessGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('code', models.CharField(max_length=50, verbose_name='Code')),
                ('is_public', models.BooleanField(default=True, help_text='If set, members from all tenants can be assigned to this group.', verbose_name='Is Public')),
            ],
            options={
                'verbose_name': 'Access Group',
                'verbose_name_plural': 'Access Groups',
                'db_table': 'accrete_access_group',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='AccessGroupMember',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'verbose_name': 'Access Group Member',
                'verbose_name_plural': 'Access Group Members',
                'db_table': 'accrete_access_group_member_rel',
                'ordering': ['member'],
            },
        ),
        migrations.CreateModel(
            name='Tenant',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('is_active', models.BooleanField(default=True, verbose_name='Active')),
            ],
            options={
                'verbose_name': 'Tenant',
                'verbose_name_plural': 'Tenants',
                'db_table': 'accrete_tenant',
                'ordering': ['-is_active', 'name'],
            },
        ),
        migrations.CreateModel(
            name='TenantAccessGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('access_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accrete.accessgroup')),
                ('tenant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accrete.tenant')),
            ],
            options={
                'verbose_name': 'Tenant Access Group',
                'verbose_name_plural': 'Tenant Access Groups',
                'db_table': 'accrete_tenant_access_group_rel',
                'ordering': ['tenant'],
            },
        ),
        migrations.AddField(
            model_name='tenant',
            name='access_groups',
            field=models.ManyToManyField(through='accrete.TenantAccessGroup', to='accrete.accessgroup'),
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, help_text='150 characters or fewer.Letters, digits and @/./+/-/_ only.', max_length=150, null=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='Name')),
                ('is_active', models.BooleanField(default=True, verbose_name='Active')),
                ('access_groups', models.ManyToManyField(through='accrete.AccessGroupMember', to='accrete.accessgroup')),
                ('tenant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='members', to='accrete.tenant')),
            ],
            options={
                'verbose_name': 'Member',
                'verbose_name_plural': 'Members',
                'db_table': 'accrete_member',
                'ordering': ['tenant', 'user'],
            },
        ),
    ]



import datetime
from decimal import Decimal
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _


class Filter:

    query_relation_depth = 4
    query_float_fields = ['DecimalField', 'FloatField']
    query_char_fields = ['CharField', 'TextField']
    query_boolean_fields = ['BooleanField']
    query_date_fields = ['DateTimeField', 'DateField']
    query_int_fields = [
        'IntegerField', 'AutoField', 'BigAutoField', 'PositiveSmallIntegerField'
    ]

    LABEL_EXACT = _('Equals')
    LABEL_EXACT_NOT = _('Equals Not')
    LABEL_ICONTAINS = _('Contains')
    LABEL_ICONTAINS_NOT = _('Contains Not')
    LABEL_GTE = _('Greater or Equal')
    LABEL_LTE = _('Less or Equal')
    LABEL_TRUE = _('True')
    LABEL_FALSE = _('False')
    LABEL_SET = _('Is Set')
    LABEL_NOT_SET = _('Is Not Set')

    def __init__(self, model, query_relation_depth: int = 4):
        self.model = model
        self.query_relation_depth = query_relation_depth

    def get_choice_char_query_term(self, label, param, choices):
        choices = [(choice[0], str(choice[1])) for choice in choices]
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_EXACT),
                    'data_type': 'selection',
                    'choices': choices,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_EXACT_NOT),
                    'data_type': 'selection',
                    'choices': choices,
                    'invert': True,
                    'param': 'exact'
                }
            ]
        }

    def get_char_query_term(self, label, param):
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_ICONTAINS),
                    'data_type': 'text',
                    'param': 'icontains'
                },
                {
                    'label': str(self.LABEL_ICONTAINS_NOT),
                    'data_type': 'text',
                    'invert': True,
                    'param': 'icontains'
                },
                {
                    'label': str(self.LABEL_EXACT),
                    'data_type': 'text',
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_EXACT_NOT),
                    'data_type': 'text',
                    'invert': True,
                    'param': 'exact'
                }
            ]
        }

    def get_float_query_term(self, label, param, step):
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_EXACT),
                    'data_type': 'number',
                    'step': step,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_EXACT_NOT),
                    'data_type': 'number',
                    'step': step,
                    'invert': True,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_GTE),
                    'data_type': 'number',
                    'step': step,
                    'param': 'gte'
                },
                {
                    'label': str(self.LABEL_LTE),
                    'data_type': 'number',
                    'step': step,
                    'param': 'lte'
                }
            ]
        }

    def get_int_query_term(self, label, param):
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_EXACT),
                    'data_type': 'number',
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_EXACT_NOT),
                    'data_type': 'number',
                    'invert': True,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_GTE),
                    'data_type': 'number',
                    'param': 'gte'
                },
                {
                    'label': str(self.LABEL_LTE),
                    'data_type': 'number',
                    'param': 'lte'
                }
            ]
        }

    def get_boolean_query_term(self, label, param):
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_TRUE),
                    'data_type': 'bool',
                    'value': True,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_FALSE),
                    'data_type': 'bool',
                    'value': False,
                    'param': 'exact'
                }
            ]
        }

    def get_datetime_query_term(self, label, param):
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_EXACT),
                    'data_type': 'datetime-local',
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_EXACT_NOT),
                    'data_type': 'datetime-local',
                    'invert': True,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_GTE),
                    'data_type': 'datetime-local',
                    'param': 'gte'
                },
                {
                    'label': str(self.LABEL_LTE),
                    'data_type': 'datetime-local',
                    'param': 'lte'
                }
            ]
        }

    def get_date_query_term(self, label, param):
        return {
            'label': str(label),
            'param': param,
            'params': [
                {
                    'label': str(self.LABEL_EXACT),
                    'data_type': 'date',
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_EXACT_NOT),
                    'data_type': 'date',
                    'invert': True,
                    'param': 'exact'
                },
                {
                    'label': str(self.LABEL_GTE),
                    'data_type': 'date',
                    'param': 'gte'
                },
                {
                    'label': str(self.LABEL_LTE),
                    'data_type': 'date',
                    'param': 'lte'
                }
            ]
        }

    def get_null_params(self):
        return [
            {
                'label': str(self.LABEL_SET),
                'data_type': 'bool',
                'value': False,
                'param': 'isnull'
            },
            {
                'label': str(self.LABEL_NOT_SET),
                'data_type': 'bool',
                'value': True,
                'param': 'isnull'
            }
        ]

    @staticmethod
    def cast_decimal_places_to_step(decimal_places):
        if not decimal_places or decimal_places < 1:
            return '1'
        zero_count = decimal_places - 1
        return f'0.{"0" * zero_count}1'

    def get_query_term(self, field):
        internal_type = field.get_internal_type()
        label = str(field.verbose_name)
        param = str(field.name)
        if internal_type in self.query_char_fields:
            if field.choices:
                return self.get_choice_char_query_term(label, param, field.choices)
            return self.get_char_query_term(label, param)
        elif internal_type in self.query_float_fields:
            return self.get_float_query_term(
                label, param, self.cast_decimal_places_to_step(field.decimal_places or 0)
            )
        elif internal_type in self.query_int_fields:
            return self.get_int_query_term(label, param)
        elif internal_type in self.query_boolean_fields:
            return self.get_boolean_query_term(label, param)
        elif internal_type in self.query_date_fields:
            return self.get_date_query_term(label, param)

    def get_annotation_term(self, model, annotation):
        name = '_a_' + annotation['name']
        label = annotation.get('label', name)
        return_type = annotation['type']
        if return_type == str:
            return self.get_char_query_term(label, name)
        elif return_type == int:
            return self.get_int_query_term(label, name)
        elif return_type in [float, Decimal]:
            return self.get_float_query_term(label, name, annotation.get('step', 0))
        elif return_type == bool:
            return self.get_boolean_query_term(label, name)
        elif return_type == datetime.datetime:
            return self.get_datetime_query_term(label, name)
        elif return_type == datetime.date:
            return self.get_date_query_term(label, name)

    def get_relation_query_terms(self, model, path):
        terms = []
        filter_exclude = getattr(model, 'filter_exclude', [])
        filter_exclude.append('tenant')
        fields = filter(
            lambda x: x.is_relation and x.name not in filter_exclude,
            sorted(model._meta.get_fields(), key=lambda x: x.name.lower())
        )
        cpath = path.copy()
        for field in fields:
            if field.related_model not in cpath:
                cpath.append(field.related_model)
                try:
                    label = field.verbose_name
                except AttributeError:
                    label = field.related_model._meta.verbose_name
                rel_term = {
                    'label': str(label),
                    'param': field.name,
                    'params': self._get_query_terms(field.related_model, cpath)
                }
                rel_term['params'] = self.get_null_params() + rel_term['params']
                if rel_term['params']:
                    terms.append(rel_term)
        return terms

    def get_local_query_terms(self, model):
        terms = []
        filter_exclude = getattr(model, 'filter_exclude', [])
        fields = filter(
            lambda x: not x.is_relation and x.name not in filter_exclude,
            model._meta.get_fields()
        )
        for field in fields:
            term = self.get_query_term(field)
            if field.null is True and term.get('params'):
                term['params'].extend(self.get_null_params())
            if term is not None:
                terms.append(term)
        for annotation in getattr(model, 'annotations', []):
            terms.append(self.get_annotation_term(model, annotation))
        terms = sorted(terms, key=lambda x: x['label'])
        return terms

    def _get_query_terms(self, model=None, path=None):
        if model is None:
            model = self.model
            path = [model]
        terms = []
        terms.extend(self.get_local_query_terms(model))
        if len(path) <= self.query_relation_depth:
            terms.extend(self.get_relation_query_terms(model, path))
        return terms

    def get_query_terms(self):
        key = f'query-terms-{self.model.__module__}.{self.model.__name__}'
        terms = cache.get(key)
        if not terms:
            terms = self._get_query_terms()
            cache.set(key, terms, 60 * 15)
        return terms



import json


def load_querystring(get_params: dict) -> list:
    return json.loads(get_params.get('q', '[]'))


def build_url_params(get_params: dict, extra_params: list[str] = None) -> str:
    querystring = f'?q={get_params.get("q", "[]")}'
    if paginate_by := get_params.get('paginate_by', False):
        querystring += f'&paginate_by={paginate_by}'
    if order_by := get_params.get('order_by', False):
        querystring += f'&order_by={order_by}'
    if crumbs := get_params.get('crumbs', False):
        querystring += f'&crumbs={crumbs}'
    for param in extra_params or []:
        if value := get_params.get(param, False):
            querystring += f'&{param}={value}'
    return querystring
